using Microsoft.Win32;
using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Security.Cryptography;

namespace RsaUtil
{
	public class RsaKeyFileUtil
	{
		private RsaKeyFileUtil() { }

        #region P/Invoke
        [DllImport("Kernel32.dll", SetLastError=true)]
        static extern IntPtr GetCurrentProcess();

        [DllImport("Kernel32.dll", SetLastError=true)]
        static extern bool CloseHandle(
            IntPtr  handle);

        [DllImport("Advapi32.dll", SetLastError=true)]
        static extern bool OpenProcessToken(
            IntPtr      hProcess,
            int         desiredAccess,
            out IntPtr  pHandle);

        [DllImport("Advapi32.dll", SetLastError=true, EntryPoint="LookupAccountNameW", CharSet=CharSet.Unicode)]
        static extern bool LookupAccountName(
            string   lpSystemName,
            string   lpAccountName,
            IntPtr   Sid,
            ref int  cbSid,
            IntPtr   ReferencedDomainName,
            ref int  cchReferencedDomainName,
            out      SidNameUse peUse);

        [DllImport("Advapi32.dll", SetLastError=true)]
        static extern bool GetTokenInformation(
            IntPtr   hToken,
            int      tokenInformationClass, // TOKEN_INFORMATION_CLASS
            IntPtr   pTokenInformation,
            int      tokenInformationLength,
            out int  returnLength);

        [DllImport("Advapi32.dll", SetLastError=true, EntryPoint="ConvertSidToStringSidW")]
        static extern bool ConvertSidToStringSid(
            IntPtr      Sid,
            out IntPtr  StringSid);

        [DllImport("Kernel32.dll", SetLastError=true)]
        static extern IntPtr LocalFree(
            IntPtr hMem);

        enum SidNameUse
        {
            SidTypeUser           = 1,
            SidTypeGroup          = 2,
            SidTypeDomain         = 3,
            SidTypeAlias          = 4,
            SidTypeWellKnownGroup = 5,
            SidTypeDeletedAccount = 6,
            SidTypeInvalid        = 7,
            SidTypeUnknown        = 8,
            SidTypeComputer       = 9
        }

        const int TokenUser                 = 1;
        const int TOKEN_QUERY               = 0x0008;
        const int ERROR_INSUFFICIENT_BUFFER = 122;
        #endregion P/Invoke

        public static string GetRsaKeyContainerFilename(string keyContainerName, bool isMachineKey)
        {
            string prefix = GetRsaKeyContainerPrefix(keyContainerName).ToLower();

            string machineGuid = GetMachineGuid();

            string filename = string.Concat(prefix, "_", machineGuid);

            string path = GetRsaKeyPath(isMachineKey);

            return Path.Combine(path, filename);
        }

        public static string GetRsaKeyPath(bool isMachineKey)
        {
            string path;

            if(isMachineKey)
            {
                string appDataPath = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData);

                path = Path.Combine(appDataPath, @"Microsoft\Crypto\RSA\MachineKeys");
            }
            else
            {
                string appDataPath = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);

                path = Path.Combine(appDataPath, @"Microsoft\Crypto\RSA");

                path = Path.Combine(path, GetCurrentUserSidString());
            }

            return path;
        }

        public static string GetRsaKeyContainerPrefix(string keyContainerName)
        {
            HashAlgorithm hash = new MD5CryptoServiceProvider();

            byte[] plain = Encoding.ASCII.GetBytes(keyContainerName.ToLower() + "\0");

            byte[] hashed = hash.ComputeHash(plain);

            for(int i = 0; i < 4; i++)
            {
                Array.Reverse(hashed, i * 4, 4);
            }

            return BinaryToHexString(hashed);
        }

        public static string GetMachineGuid()
        {
            using(RegistryKey rk = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Microsoft\Cryptography"))
            {
                return rk.GetValue("MachineGuid").ToString();
            }
        }

        public static string BinaryToHexString(byte[] binary)
        {
            StringBuilder hex = new StringBuilder(binary.Length * 2);

            foreach(byte b in binary)
                hex.AppendFormat("{0:X2}", b);

            return hex.ToString();
        }

        public static string GetCurrentUserSidString()
        {
            IntPtr hToken     = IntPtr.Zero;
            IntPtr pSidAtt    = IntPtr.Zero;
            IntPtr pStringSid = IntPtr.Zero;
            int    length     = 0;

            try
            {
                if(!OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, out hToken))
                {
                    throw new ApplicationException("OpenProcessToken failed.  Error = " + Marshal.GetLastWin32Error().ToString());
                }

                GetTokenInformation(hToken, TokenUser, pSidAtt, length, out length);

                int error = Marshal.GetLastWin32Error();

                if(error != ERROR_INSUFFICIENT_BUFFER)
                {
                    throw new ApplicationException("GetTokenInformation failed.  Error = " + error.ToString());
                }

                pSidAtt = Marshal.AllocHGlobal(length);
                
                if(!GetTokenInformation(hToken, TokenUser, pSidAtt, length, out length))
                {
                    throw new ApplicationException("GetTokenInformation failed.  Error = " + Marshal.GetLastWin32Error().ToString());
                }

                IntPtr pSid = Marshal.ReadIntPtr(pSidAtt);

                if(!ConvertSidToStringSid(pSid, out pStringSid))
                {
                    throw new ApplicationException("ConvertSidToStringSid failed.  Error = " + Marshal.GetLastWin32Error().ToString());
                }

                return Marshal.PtrToStringUni(pStringSid);
            }
            finally
            {
                if(pStringSid != IntPtr.Zero)
                {
                    LocalFree(pStringSid);
                }

                if(pSidAtt != IntPtr.Zero)
                {
                    Marshal.FreeHGlobal(pSidAtt);
                }

                if(hToken != IntPtr.Zero)
                {
                    CloseHandle(hToken);
                }
            }
        }
	}
}
